/* doscan - Denial Of Service Capable Auditing of Networks
 * Copyright (C) 2003 Florian Weimer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"
#include "opt.h"

const char *opt_program = 0;

int opt_verbose = 0;
unsigned opt_port = 0;
ticks_t opt_connect_timeout = 30000;
ticks_t opt_read_timeout = 30000;
ticks_t opt_write_timeout = 30000;
unsigned opt_fd_count = 50;
unsigned opt_add_burst = 10;
ticks_t opt_add_timeout = 20;
int opt_net_errors = 0;
int opt_no_epoll = 0;
int opt_indicator = 0;
unsigned opt_banner_size = 0;
const char *opt_receive = "";
const char *opt_send = "";
const char *opt_output = "%T\t%a\t%e\t%b";
const char *opt_protocol = "tcp";
const char *opt_output_style = "sort-by-host";
